@@ -63,15 +63,24 @@ Prefix argument N means scroll forward N lines."
 	      (let ((w (vm-get-visible-buffer-window (current-buffer)))
 		    old-w-start)
 		(setq old-w-start (window-start w))
-		;; save-excurison to avoid possible buffer change
+		;; save-excursion to avoid possible buffer change
 		(save-excursion (vm-select-frame (window-frame w)))
 		(vm-raise-frame (window-frame w))
 		(vm-display nil nil '(vm-scroll-forward vm-scroll-backward)
 			    (list this-command 'reading-message))
 		(setq w (vm-get-visible-buffer-window (current-buffer)))
 		(and w (set-window-start w old-w-start))))
-	  (if (eq vm-system-state 'previewing)
-	      (vm-show-current-message))
+	  (cond ((eq vm-system-state 'previewing)
+		 (vm-show-current-message)
+		 ;; The window start marker sometimes drifts forward
+		 ;; because of something that vm-show-current-message
+		 ;; does.  In Emacs 20, replacing ASCII chars with
+		 ;; multibyte chars seems to cause it, but I _think_
+		 ;; the drift can happen in Emacs 19 and even
+		 ;; XEmacs for different reasons.  So we reset the
+		 ;; start marker here, since it is an easy fix.
+		 (let ((w (vm-get-visible-buffer-window (current-buffer))))
+		   (set-window-start w (point-min)))))
 	  (vm-howl-if-eom))
       (let ((vmp vm-message-pointer)
 	    (msg-buf (current-buffer))
@@ -106,6 +115,18 @@ Prefix argument N means scroll forward N lines."
 		(cond ((and (not (eq result 'next-message))
 			    vm-honor-page-delimiters)
 		       (vm-narrow-to-page)
+		       (goto-char (max (window-start w)
+				       (vm-text-of (car vmp))))
+		       ;; This is needed because in some cases
+		       ;; the scroll-up call in vm-howl-if-emo
+		       ;; does not signal end-of-buffer when
+		       ;; it should unless we do this.  This
+		       ;; sit-for most likely removes the need
+		       ;; for the (scroll-up 0) below, but
+		       ;; since the voodoo has worked this
+		       ;; long, it's probably best to let it
+		       ;; be.
+		       (sit-for 0)
 		       ;; This voodoo is required!  For some
 		       ;; reason the 18.52 emacs display
 		       ;; doesn't immediately reflect the
@@ -199,19 +220,15 @@ Prefix argument N means scroll forward N lines."
 	 (vm-emit-eom-blurb))))
 
 (defun vm-emit-eom-blurb ()
-  (if (vm-full-name-of (car vm-message-pointer))
-      (progn
-	(if (and (stringp vm-summary-uninteresting-senders)
-		 (string-match vm-summary-uninteresting-senders
-			       (vm-su-from (car vm-message-pointer))))
-	    (message "End of message %s to %s"
-		     (vm-number-of (car vm-message-pointer))
-		     (vm-su-to-names (car vm-message-pointer)))
-	  (message "End of message %s from %s"
-		   (vm-number-of (car vm-message-pointer))
-		   (vm-full-name-of (car vm-message-pointer)))))
-    (message "End of message %s"
-	     (vm-number-of (car vm-message-pointer)))))
+  (let ((vm-summary-uninteresting-senders-arrow "")
+	(case-fold-search nil))
+    (message (if (and (stringp vm-summary-uninteresting-senders)
+		      (string-match vm-summary-uninteresting-senders
+				    (vm-su-from (car vm-message-pointer))))
+		 "End of message %s to %s"
+	       "End of message %s from %s")
+	     (vm-number-of (car vm-message-pointer))
+	     (vm-summary-sprintf "%F" (car vm-message-pointer)))))
 
 (defun vm-scroll-backward (&optional arg)
   "Scroll backward a screenful of text.
@@ -315,6 +332,7 @@ Prefix N scrolls backward N lines."
 		  (define-key keymap "\r"
 		    (function (lambda () (interactive)
 				(vm-mouse-send-url-at-position (point)))))
+		  (set-extent-property e 'vm-button t)
 		  (set-extent-property e 'keymap keymap)
 		  (set-extent-property e 'balloon-help 'vm-url-help)
 		  (set-extent-property e 'highlight t))))
@@ -351,6 +369,7 @@ Prefix N scrolls backward N lines."
 			     (looking-at "mailto:"))
 			   'vm-menu-popup-mailto-url-browser-menu
 			 'vm-menu-popup-url-browser-menu)))
+		  (overlay-put o 'vm-button t)
 		  (overlay-put o 'mouse-face 'highlight)
 		  (setq keymap (nconc keymap (current-local-map)))
 		  (if vm-popup-menu-on-mouse-3
@@ -576,14 +595,28 @@ Use mouse button 3 to choose a Web browser for the URL."
    (if (and vm-display-using-mime
 	    vm-auto-decode-mime-messages
 	    vm-mime-decode-for-preview
+	    vm-preview-lines
 	    (if vm-mail-buffer
 		(not (vm-buffer-variable-value vm-mail-buffer
 					       'vm-mime-decoded))
 	      (not vm-mime-decoded))
 	    (not (vm-mime-plain-message-p (car vm-message-pointer))))
-       (progn
+       ;; restrict the things that are auto-displayed, since
+       ;; decode-for-preview is meant to allow a numeric
+       ;; vm-preview-lines to be useful in the face of multipart
+       ;; messages.
+       (let ((vm-auto-displayed-mime-content-type-exceptions
+	      (cons "message/external-body" vm-auto-displayed-mime-content-type-exceptions))
+	     (vm-mime-external-content-types-alist nil))
 	 (condition-case data
-	     (vm-decode-mime-message)
+	     (progn
+	       (vm-decode-mime-message)
+	       ;; reset vm-mime-decoded so that when the user
+	       ;; opens the message completely, the full MIME
+	       ;; display will happen.
+	       (and vm-mail-buffer
+		    (vm-set-buffer-variable vm-mail-buffer
+					    'vm-mime-decoded nil)))
 	   (vm-mime-error (vm-set-mime-layout-of (car vm-message-pointer)
 						 (car (cdr data)))
 			  (message "%s" (car (cdr data)))))
@@ -706,15 +739,41 @@ Use mouse button 3 to choose a Web browser for the URL."
 			  (vm-text-end-of (car vm-message-pointer))))))
 
 (defun vm-narrow-to-page ()
+  (cond (vm-fsfemacs-p
+	 (if (not (and vm-page-end-overlay
+		       (overlay-buffer vm-page-end-overlay)))
+	     (let ((g vm-page-continuation-glyph))
+	       (setq vm-page-end-overlay (make-overlay (point) (point)))
+	       (vm-set-extent-property vm-page-end-overlay 'vm-glyph g)
+	       (vm-set-extent-property vm-page-end-overlay 'before-string g)
+	       (overlay-put vm-page-end-overlay 'evaporate nil))))
+	(vm-xemacs-p
+	 (if (not (and vm-page-end-overlay
+		       (extent-end-position vm-page-end-overlay)))
+	     (let ((g vm-page-continuation-glyph))
+	       (cond ((not (glyphp g))
+		      (setq g (make-glyph g))
+		      (set-glyph-face g 'italic)))
+	       (setq vm-page-end-overlay (make-extent (point) (point)))
+	       (vm-set-extent-property vm-page-end-overlay 'vm-glyph g)
+	       (vm-set-extent-property vm-page-end-overlay 'begin-glyph g)
+	       (set-extent-property vm-page-end-overlay 'detachable nil)))))
   (save-excursion
-    (let (min max (omin (point-min)) (omax (point-max)))
+    (let (min max (e vm-page-end-overlay))
       (if (or (bolp) (not (save-excursion
 			    (beginning-of-line)
 			    (looking-at page-delimiter))))
 	  (forward-page -1))
       (setq min (point))
       (forward-page 1)
-      (beginning-of-line)
+      (if (not (eobp))
+	  (beginning-of-line))
+      (cond ((/= (point) (vm-text-end-of (car vm-message-pointer)))
+	     (vm-set-extent-property e vm-begin-glyph-property
+				     (vm-extent-property e 'vm-glyph))
+	     (vm-set-extent-endpoints e (point) (point)))
+	    (t
+	     (vm-set-extent-property e vm-begin-glyph-property nil)))
       (setq max (point))
       (narrow-to-region min max))))
 
@@ -732,13 +791,14 @@ Use mouse button 3 to choose a Web browser for the URL."
   (push-mark)
   (vm-display (current-buffer) t '(vm-beginning-of-message)
 	      '(vm-beginning-of-message reading-message))
-  (let ((osw (selected-window)))
-    (unwind-protect
-	(progn
-	  (select-window (vm-get-visible-buffer-window (current-buffer)))
-	  (goto-char (point-min)))
-      (if (not (eq osw (selected-window)))
-	  (select-window osw))))
+  (vm-save-buffer-excursion
+    (let ((osw (selected-window)))
+      (unwind-protect
+	  (progn
+	    (select-window (vm-get-visible-buffer-window (current-buffer)))
+	    (goto-char (point-min)))
+	(if (not (eq osw (selected-window)))
+	    (select-window osw)))))
   (if vm-honor-page-delimiters
       (vm-narrow-to-page)))
 
@@ -760,12 +820,102 @@ as necessary."
   (push-mark)
   (vm-display (current-buffer) t '(vm-end-of-message)
 	      '(vm-end-of-message reading-message))
-  (let ((osw (selected-window)))
-    (unwind-protect
-	(progn
-	  (select-window (vm-get-visible-buffer-window (current-buffer)))
-	  (goto-char (point-max)))
-      (if (not (eq osw (selected-window)))
-	  (select-window osw))))
+  (vm-save-buffer-excursion
+    (let ((osw (selected-window)))
+      (unwind-protect
+	  (progn
+	    (select-window (vm-get-visible-buffer-window (current-buffer)))
+	    (goto-char (point-max)))
+	(if (not (eq osw (selected-window)))
+	    (select-window osw)))))
   (if vm-honor-page-delimiters
       (vm-narrow-to-page)))
+
+(defun vm-move-to-next-button (count)
+  "Moves to the next button in the current message.
+Prefix argument N means move to the Nth next button.
+Negavite N means move to the Nth previous button.
+If there is no next button, an error is signaled and point is not moved.
+
+A button is a highlighted region of text where pressing RETURN
+will produce an action.  If the message is being previewed, it is
+exposed and marked as read."
+  (interactive "p")
+  (vm-follow-summary-cursor)
+  (vm-select-folder-buffer)
+  (vm-check-for-killed-summary)
+  (vm-check-for-killed-presentation)
+  (vm-error-if-folder-empty)
+  (and vm-presentation-buffer
+       (set-buffer vm-presentation-buffer))
+  (if (eq vm-system-state 'previewing)
+      (vm-show-current-message))
+  (setq vm-system-state 'reading)
+  (vm-widen-page)
+  (vm-display (current-buffer) t '(vm-move-to-next-button)
+	      '(vm-move-to-next-button reading-message))
+  (select-window (vm-get-visible-buffer-window (current-buffer)))
+  (unwind-protect
+      (vm-move-to-xxxx-button (vm-abs count) (>= count 0))
+    (if vm-honor-page-delimiters
+	(vm-narrow-to-page))))
+
+(defun vm-move-to-previous-button (count)
+  "Moves to the previous button in the current message.
+Prefix argument N means move to the Nth previous button.
+Negavite N means move to the Nth next button.
+If there is no previous button, an error is signaled and point is not moved.
+
+A button is a highlighted region of text where pressing RETURN
+will produce an action.  If the message is being previewed, it is
+exposed and marked as read."
+  (interactive "p")
+  (vm-follow-summary-cursor)
+  (vm-select-folder-buffer)
+  (vm-check-for-killed-summary)
+  (vm-check-for-killed-presentation)
+  (vm-error-if-folder-empty)
+  (and vm-presentation-buffer
+       (set-buffer vm-presentation-buffer))
+  (if (eq vm-system-state 'previewing)
+      (vm-show-current-message))
+  (setq vm-system-state 'reading)
+  (vm-widen-page)
+  (vm-display (current-buffer) t '(vm-move-to-previous-button)
+	      '(vm-move-to-previous-button reading-message))
+  (select-window (vm-get-visible-buffer-window (current-buffer)))
+  (unwind-protect
+      (vm-move-to-xxxx-button (vm-abs count) (< count 0))
+    (if vm-honor-page-delimiters
+	(vm-narrow-to-page))))
+
+(defun vm-move-to-xxxx-button (count next)
+  (let ((old-point (point))
+	(endp (if next 'eobp 'bobp))
+	(extent-end-position (if vm-xemacs-p
+				 (if next
+				     'extent-end-position
+				   'extent-start-position)
+			       (if next
+				   'overlay-end
+				 'overlay-start)))
+	(next-extent-change (if vm-xemacs-p
+				(if next
+				    'next-extent-change
+				  'previous-extent-change)
+			      (if next
+				  'next-overlay-change
+				'previous-overlay-change)))
+	e)
+    (while (and (> count 0) (not (funcall endp)))
+      (goto-char (funcall next-extent-change (+ (point) (if next 0 -1))))
+      (setq e (vm-extent-at (point)))
+      (if e
+	  (progn
+	    (if (vm-extent-property e 'vm-button)
+		(vm-decrement count))
+	    (goto-char (funcall extent-end-position e)))))
+    (if e
+	(goto-char (vm-extent-start-position e))
+      (goto-char old-point)
+      (error "No more buttons"))))
