@@ -137,7 +137,8 @@ The saved messages are flagged as `filed'."
 		      (format "Save message %s in folder %s? "
 			      (vm-number-of (car vm-message-pointer))
 			      auto-folder)))
-		 (let ((vm-delete-after-saving vm-delete-after-archiving))
+		 (let ((vm-delete-after-saving vm-delete-after-archiving)
+		       (last-command 'vm-auto-archive-messages))
 		   (vm-save-message auto-folder)
 		   (vm-increment archived)
 		   (message "%d archived, still working..." archived)))
@@ -513,7 +514,7 @@ Output, if any, is displayed.  The message is not altered."
 	(display-buffer buffer))))
 
 (defun vm-print-message (&optional count)
-  "Print the current message.
+  "Print the current message
 Prefix arg N means print the current message and the next N - 1 messages.
 Prefix arg -N means print the current message and the previous N - 1 messages.
 
