@@ -1,5 +1,5 @@
 ;;; Toolbar related functions and commands
-;;; Copyright (C) 1995-1997 Kyle E. Jones
+;;; Copyright (C) 1995-1997, 2000 Kyle E. Jones
 ;;;
 ;;; This program is free software; you can redistribute it and/or modify
 ;;; it under the terms of the GNU General Public License as published by
@@ -149,8 +149,11 @@ s-expression like this one in your .vm file:
 (or (fboundp 'vm-toolbar-decode-mime-command)
     (fset 'vm-toolbar-decode-mime-command 'vm-decode-mime-message))
 
-(defvar vm-toolbar-delete-icon nil)
-
+;; The values of these two are used by the FSF Emacs toolbar
+;; code.  The values don't matter as long as they are different
+;; (as compared with eq).  Under XEmacs these values are ignored
+;; and overwritten.
+(defvar vm-toolbar-delete-icon t)
 (defvar vm-toolbar-undelete-icon nil)
 
 (defvar vm-toolbar-delete/undelete-button
@@ -173,9 +176,13 @@ s-expression like this one in your .vm file:
    (vm-toolbar-can-help-p)
    "Don't Panic.\n
 VM uses this button to offer help if you're in trouble.
-Under normal circumstances, this button runs `vm-help'.\n
+Under normal circumstances, this button runs `vm-help'.
 If the current folder looks out-of-date relative to its auto-save
-file then this button will run `recover-file'."])
+file then this button will run `recover-file'
+If there is mail waiting in one of the spool files associated
+with the current folder, this button will run `vm-get-new-mail'.
+If the current message needs to be MIME decoded then this button
+will run 'vm-decode-mime-message'."])
 
 (defvar vm-toolbar-helper-command nil)
 (make-variable-buffer-local 'vm-toolbar-helper-command)
@@ -202,7 +209,7 @@ s-expression like this one in your .vm file:
   (condition-case nil
       (save-excursion
 	(vm-check-for-killed-folder)
-	(vm-select-folder-buffer)
+	(vm-select-folder-buffer-if-possible)
 	vm-message-list)
     (error nil)))
 
@@ -223,7 +230,7 @@ s-expression like this one in your .vm file:
   (condition-case nil
       (save-excursion
 	(vm-check-for-killed-folder)
-	(vm-select-folder-buffer)
+	(vm-select-folder-buffer-if-possible)
 	(and vm-message-pointer
 	     (vm-auto-select-folder vm-message-pointer vm-auto-folder-alist)))
     (error nil)))
@@ -316,50 +323,58 @@ s-expression like this one in your .vm file:
 						   vm-toolbar)))))
 
 (defun vm-toolbar-install-toolbar ()
-  (vm-toolbar-initialize)
-  (let ((height (+ 4 (glyph-height (car vm-toolbar-help-icon))))
-	(width (+ 4 (glyph-width (car vm-toolbar-help-icon))))
-	(frame (selected-frame))
-	(buffer (current-buffer))
-	(tag-set '(win))
-	(myframe (vm-created-this-frame-p))
-	toolbar )
-    ;; glyph-width and glyph-height return 0 at startup sometimes
-    ;; use reasonable values if they fail.
-    (if (= width 4)
-	(setq width 68))
-    (if (= height 4)
-	(setq height 46))
-    ;; honor user setting of vm-toolbar if they are daring enough
-    ;; to set it.
-    (if vm-toolbar
-	(setq toolbar vm-toolbar)
-      (setq toolbar (vm-toolbar-make-toolbar-spec)
-	    vm-toolbar toolbar))
-    (cond ((eq vm-toolbar-orientation 'right)
-	   (setq vm-toolbar-specifier right-toolbar)
-	   (if myframe
-	       (set-specifier right-toolbar toolbar frame tag-set))
-	   (set-specifier right-toolbar toolbar buffer)
-	   (set-specifier right-toolbar-width width frame tag-set))
-	  ((eq vm-toolbar-orientation 'left)
-	   (setq vm-toolbar-specifier left-toolbar)
-	   (if myframe
-	       (set-specifier left-toolbar toolbar frame tag-set))
-	   (set-specifier left-toolbar toolbar buffer)
-	   (set-specifier left-toolbar-width width frame tag-set))
-	  ((eq vm-toolbar-orientation 'bottom)
-	   (setq vm-toolbar-specifier bottom-toolbar)
-	   (if myframe
-	       (set-specifier bottom-toolbar toolbar frame tag-set))
-	   (set-specifier bottom-toolbar toolbar buffer)
-	   (set-specifier bottom-toolbar-height height frame tag-set))
-	  (t
-	   (setq vm-toolbar-specifier top-toolbar)
-	   (if myframe
-	       (set-specifier top-toolbar toolbar frame tag-set))
-	   (set-specifier top-toolbar toolbar buffer)
-	   (set-specifier top-toolbar-height height frame tag-set)))))
+  (if vm-fsfemacs-p
+      (if (not vm-fsfemacs-toolbar-installed-p)
+	  (vm-toolbar-fsfemacs-install-toolbar))
+    (if (not (and (stringp vm-toolbar-pixmap-directory)
+		  (file-directory-p vm-toolbar-pixmap-directory)))
+	(progn
+	  (message "Bad toolbar pixmap directory, can't setup toolbar.")
+	  (sit-for 2))
+      (vm-toolbar-initialize)
+      (let ((height (+ 4 (glyph-height (car vm-toolbar-help-icon))))
+	    (width (+ 4 (glyph-width (car vm-toolbar-help-icon))))
+	    (frame (selected-frame))
+	    (buffer (current-buffer))
+	    (tag-set '(win))
+	    (myframe (vm-created-this-frame-p))
+	    toolbar )
+	;; glyph-width and glyph-height return 0 at startup sometimes
+	;; use reasonable values if they fail.
+	(if (= width 4)
+	    (setq width 68))
+	(if (= height 4)
+	    (setq height 46))
+	;; honor user setting of vm-toolbar if they are daring enough
+	;; to set it.
+	(if vm-toolbar
+	    (setq toolbar vm-toolbar)
+	  (setq toolbar (vm-toolbar-make-toolbar-spec)
+		vm-toolbar toolbar))
+	(cond ((eq vm-toolbar-orientation 'right)
+	       (setq vm-toolbar-specifier right-toolbar)
+	       (if myframe
+		   (set-specifier right-toolbar toolbar frame tag-set))
+	       (set-specifier right-toolbar toolbar buffer)
+	       (set-specifier right-toolbar-width width frame tag-set))
+	      ((eq vm-toolbar-orientation 'left)
+	       (setq vm-toolbar-specifier left-toolbar)
+	       (if myframe
+		   (set-specifier left-toolbar toolbar frame tag-set))
+	       (set-specifier left-toolbar toolbar buffer)
+	       (set-specifier left-toolbar-width width frame tag-set))
+	      ((eq vm-toolbar-orientation 'bottom)
+	       (setq vm-toolbar-specifier bottom-toolbar)
+	       (if myframe
+		   (set-specifier bottom-toolbar toolbar frame tag-set))
+	       (set-specifier bottom-toolbar toolbar buffer)
+	       (set-specifier bottom-toolbar-height height frame tag-set))
+	      (t
+	       (setq vm-toolbar-specifier top-toolbar)
+	       (if myframe
+		   (set-specifier top-toolbar toolbar frame tag-set))
+	       (set-specifier top-toolbar toolbar buffer)
+	       (set-specifier top-toolbar-height height frame tag-set)))))))
 
 (defun vm-toolbar-make-toolbar-spec ()
   (let ((button-alist '(
@@ -403,6 +418,7 @@ s-expression like this one in your .vm file:
   (require 'vm-save)
   (require 'vm-summary)
   (cond
+   (vm-fsfemacs-p nil)
    ((null vm-toolbar-help-icon)
     (let ((tuples
 	   (if (featurep 'xpm)
@@ -454,22 +470,213 @@ s-expression like this one in your .vm file:
  (vm-toolbar-recover-icon "recover-up.xbm" "recover-dn.xbm" "recover-xx.xpm")
 	   )))
 	  tuple files var)
-      (if (not (file-directory-p vm-toolbar-pixmap-directory))
-	  (error "Bad toolbar pixmap directory: %s"
-		 vm-toolbar-pixmap-directory)
-	(while tuples
-	  (setq tuple (car tuples)
-		var (car tuple)
-		files (cdr tuple))
-	  (set var (mapcar
-		    (function
-		     (lambda (f)
-		       (make-glyph
-			(expand-file-name f vm-toolbar-pixmap-directory))))
-		    files))
-	  (setq tuples (cdr tuples)))))))
+      (while tuples
+	(setq tuple (car tuples)
+	      var (car tuple)
+	      files (cdr tuple))
+	(set var (mapcar
+		  (function
+		   (lambda (f)
+		     (make-glyph
+		      (expand-file-name f vm-toolbar-pixmap-directory))))
+		  files))
+	(setq tuples (cdr tuples))))))
   (setq vm-toolbar-delete/undelete-icon vm-toolbar-delete-icon)
   (setq-default vm-toolbar-delete/undelete-icon vm-toolbar-delete-icon)
   (setq vm-toolbar-helper-command 'vm-help)
   (setq vm-toolbar-helper-icon vm-toolbar-help-icon)
   (setq-default vm-toolbar-helper-icon vm-toolbar-help-icon))
+
+(defun vm-toolbar-fsfemacs-install-toolbar ()
+  (let ((button-list (reverse vm-use-toolbar))
+	(dir vm-toolbar-pixmap-directory)
+	(extension (if (and (display-color-p)
+			    (image-type-available-p 'xpm))
+		       "xpm"
+		     "xbm"))
+	item t-spec sym name images)
+    (defvar tool-bar-map)
+    ;; hide the toolbar entries that are in the global keymap so
+    ;; VM has full control of the toolbar in its buffers.
+    (if (and (boundp 'tool-bar-map)
+	     (consp tool-bar-map))
+	(let ((map (cdr tool-bar-map))
+	      (v [tool-bar x]))
+	  (while map
+	    (aset v 1 (car (car map)))
+	    (define-key vm-mode-map v 'undefined)
+	    (setq map (cdr map)))))
+    (while button-list
+      (setq sym (car button-list))
+      (cond ((null sym)
+	     ;; can't do flushright in FSF Emacs
+	     t)
+	    ((integerp sym)
+	     ;; can't do separators in FSF Emacs
+	     t)
+	    ((memq sym '(autofile compose file getmail
+			 mime next previous print quit reply visit))
+	     (setq t-spec (symbol-value
+			   (intern (format "vm-toolbar-%s-button"
+					   (if (eq sym 'mime)
+					       'decode-mime
+					     sym)))))
+	     (if (and (eq sym 'mime) (string= extension "xpm"))
+		 (setq name "mime-colorful")
+	       (setq name (symbol-name sym)))
+	     (setq images (vm-toolbar-make-fsfemacs-toolbar-image-spec
+			   name extension dir))
+	     (setq item
+		   (list 'menu-item
+			 (aref t-spec 3)
+			 (aref t-spec 1)
+			 ':enable (aref t-spec 2)
+			 ':button '(:toggle nil)
+			 ':image images))
+	     (define-key vm-mode-map (vector 'tool-bar sym) item))
+	    ((eq sym 'delete/undelete)
+	     (setq t-spec vm-toolbar-delete/undelete-button)
+	     (setq name "delete")
+	     (setq images (vm-toolbar-make-fsfemacs-toolbar-image-spec
+			   name extension dir))
+	     (setq item
+		   (list 'menu-item
+			 (aref t-spec 3)
+			 (aref t-spec 1)
+			 ':visible '(eq vm-toolbar-delete/undelete-icon
+					vm-toolbar-delete-icon)
+			 ':enable (aref t-spec 2)
+			 ':button '(:toggle nil)
+			 ':image images))
+	     (define-key vm-mode-map (vector 'tool-bar 'delete) item)
+	     (setq name "undelete")
+	     (setq images (vm-toolbar-make-fsfemacs-toolbar-image-spec
+			   name extension dir))
+	     (setq item
+		   (list 'menu-item
+			 (aref t-spec 3)
+			 (aref t-spec 1)
+			 ':visible '(eq vm-toolbar-delete/undelete-icon
+					vm-toolbar-undelete-icon)
+			 ':enable (aref t-spec 2)
+			 ':button '(:toggle nil)
+			 ':image images))
+	     (define-key vm-mode-map (vector 'tool-bar 'undelete) item))
+	    ((eq sym 'help)
+	     (setq t-spec vm-toolbar-help-button)
+	     (setq name "help")
+	     (setq images (vm-toolbar-make-fsfemacs-toolbar-image-spec
+			   name extension dir))
+	     (setq item
+		   (list 'menu-item
+			 (aref t-spec 3)
+			 (aref t-spec 1)
+			 ':visible '(eq vm-toolbar-helper-command 'vm-help)
+			 ':enable (aref t-spec 2)
+			 ':button '(:toggle nil)
+			 ':image images))
+	     (define-key vm-mode-map (vector 'tool-bar 'help-help) item)
+	     (setq name "recover")
+	     (setq images (vm-toolbar-make-fsfemacs-toolbar-image-spec
+			   name extension dir))
+	     (setq item
+		   (list 'menu-item
+			 (aref t-spec 3)
+			 (aref t-spec 1)
+			 ':visible '(eq vm-toolbar-helper-command
+					'recover-file)
+			 ':enable (aref t-spec 2)
+			 ':button '(:toggle nil)
+			 ':image images))
+	     (define-key vm-mode-map (vector 'tool-bar 'help-recover) item)
+	     (setq name "getmail")
+	     (setq images (vm-toolbar-make-fsfemacs-toolbar-image-spec
+			   name extension dir))
+	     (setq item
+		   (list 'menu-item
+			 (aref t-spec 3)
+			 (aref t-spec 1)
+			 ':visible '(eq vm-toolbar-helper-command
+					'vm-get-new-mail)
+			 ':enable (aref t-spec 2)
+			 ':button '(:toggle nil)
+			 ':image images))
+	     (define-key vm-mode-map (vector 'tool-bar 'help-getmail) item)
+	     (if (string= extension "xpm")
+		 (setq name "mime-colorful")
+	       (setq name "mime"))
+	     (setq images (vm-toolbar-make-fsfemacs-toolbar-image-spec
+			   name extension dir))
+	     (setq item
+		   (list 'menu-item
+			 (aref t-spec 3)
+			 (aref t-spec 1)
+			 ':visible '(eq vm-toolbar-helper-command
+					'vm-decode-mime-message)
+			 ':enable (aref t-spec 2)
+			 ':button '(:toggle nil)
+			 ':image images))
+	     (define-key vm-mode-map (vector 'tool-bar 'help-mime) item)))
+      (setq button-list (cdr button-list))))
+  (setq vm-fsfemacs-toolbar-installed-p t))
+
+(defun vm-toolbar-make-fsfemacs-toolbar-image-spec (name extension dir)
+  (if (string= extension "xpm")
+      (vector
+       (list 'image
+	     ':type (intern extension)
+	     ':mask 'heuristic
+	     ':file (expand-file-name
+		     (format "%s-dn.%s"
+			     name extension)
+		     dir))
+       (list 'image
+	     ':type (intern extension)
+	     ':mask 'heuristic
+	     ':file (expand-file-name
+		     (format "%s-up.%s"
+			     name extension)
+		     dir))
+       (list 'image
+	     ':type (intern extension)
+	     ':mask 'heuristic
+	     ':file (expand-file-name
+		     (format "%s-dn.%s"
+			     name extension)
+		     dir))
+       (list 'image
+	     ':type (intern extension)
+	     ':mask 'heuristic
+	     ':file (expand-file-name
+		     (format "%s-dn.%s"
+			     name extension)
+		     dir)))
+    (vector
+     (list 'image
+	   ':type (intern extension)
+	   ':mask 'heuristic
+	   ':file (expand-file-name
+		   (format "%s-dn.%s"
+			   name extension)
+		   dir))
+     (list 'image
+	   ':type (intern extension)
+	   ':mask 'heuristic
+	   ':file (expand-file-name
+		   (format "%s-up.%s"
+			   name extension)
+		   dir))
+     (list 'image
+	   ':type (intern extension)
+	   ':mask 'heuristic
+	   ':file (expand-file-name
+		   (format "%s-xx.%s"
+			   name extension)
+		   dir))
+     (list 'image
+	   ':type (intern extension)
+	   ':mask 'heuristic
+	   ':file (expand-file-name
+		   (format "%s-xx.%s"
+			   name extension)
+		   dir)))))
